# Yanked from Rails

# Copyright (c) 2004-2009 David Heinemeier Hansson
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

module Tramp
  module Quoting
    def quote_column_name(name)
      @quoted_column_names[name] ||= "`#{name}`"
    end

    def quote_table_name(name)
      @quoted_table_names[name] ||= quote_column_name(name).gsub('.', '`.`')
    end

    def quote(value, column = nil)
      if value.kind_of?(String) && column && column.type == :binary && column.class.respond_to?(:string_to_binary)
        s = value.unpack("H*")[0]
        "x'#{s}'"
      elsif value.kind_of?(BigDecimal)
        value.to_s("F")
      else
        super
      end
    end

    def quote(value, column = nil)
      # records are quoted as their primary key
      return value.quoted_id if value.respond_to?(:quoted_id)

      case value
        when String, ActiveSupport::Multibyte::Chars
          value = value.to_s
          if column && column.type == :binary && column.class.respond_to?(:string_to_binary)
            "#{quoted_string_prefix}'#{quote_string(column.class.string_to_binary(value))}'" # ' (for ruby-mode)
          elsif column && [:integer, :float].include?(column.type)
            value = column.type == :integer ? value.to_i : value.to_f
            value.to_s
          else
            "#{quoted_string_prefix}'#{quote_string(value)}'" # ' (for ruby-mode)
          end
        when NilClass                 then "NULL"
        when TrueClass                then (column && column.type == :integer ? '1' : quoted_true)
        when FalseClass               then (column && column.type == :integer ? '0' : quoted_false)
        when Float, Fixnum, Bignum    then value.to_s
        # BigDecimals need to be output in a non-normalized form and quoted.
        when BigDecimal               then value.to_s('F')
        else
          if value.acts_like?(:date) || value.acts_like?(:time)
            "'#{quoted_date(value)}'"
          else
            "#{quoted_string_prefix}'#{quote_string(value.to_yaml)}'"
          end
      end
    end

    # Quotes a string, escaping any ' (single quote) and \ (backslash)
    # characters.
    def quote_string(s)
      s.gsub(/\\/, '\&\&').gsub(/'/, "''") # ' (for ruby-mode)
    end

    def quoted_true
      "'t'"
    end

    def quoted_false
      "'f'"
    end

    def quoted_date(value)
      if value.acts_like?(:time)
        zone_conversion_method = ActiveRecord::Base.default_timezone == :utc ? :getutc : :getlocal
        value.respond_to?(zone_conversion_method) ? value.send(zone_conversion_method) : value
      else
        value
      end.to_s(:db)
    end

    def quoted_string_prefix
      ''
    end
  end
end
